/* Phyperoid 1.1 - an enhanced shoot-the-asteroids game.
 * Copyright (C) 1991 Edward Hutchins.
 * PalmOS changes copyright (C) 2002 Russell Marks.
 *
 * draw.c - (hopefully) fast drawing to offscreen buffer.
 *
 * Note that this all requires OS 2.0 or better, so that should
 * be checked for before you ever touch this stuff. :-)
 */

/* Provides a 160x160 1-bit mono screen for all screen types it supports:
 *
 * - 160x160 1-bit (and 2/4-bit if they can switch into 1-bit mode)
 * - 160x160 8-bit colour (and 12?/16-bit if they can switch)
 *
 * On other screens it'll resort to using OS calls, which is MUCH slower.
 */

/* the draw_line* routines are hacked versions of:
 *
 * digline: draw digital line from (x1,y1) to (x2,y2),
 * calling a user-supplied procedure at each pixel.
 * Does no clipping.  Uses Bresenham's algorithm.
 *
 * Paul Heckbert	3 Sep 85
 *
 * ...which is from "Graphics Gems", Academic Press, 1990.
 */

#include <PalmOS.h>
#include <string.h>

#include "draw.h"
#include "misc.h"
#include "hyperoid.h"


void (*draw_pixel)(int,int);
void (*draw_line)(int,int,int,int);
void (*draw_bmp)(int,int,int,int,unsigned char *);

static int featureset30=0,featureset35=0;
static IndexedColorType cols_8bit[8];
static IndexedColorType white_8bit,black_8bit;
static int buffersize=0;

/* this is exported to allow for warning alerts, etc. */
int draw_is_using_os=0;

static UInt32 scrwidth,scrheight,scrdepth;
static Boolean screncol;

static WinHandle offscreenwin,currentwin;
static RectangleType oswbounds;
static unsigned char *scrnbmp;

static int cols_changed=0;
static IndexedColorType old_back,old_fore;


/* absolute value of a */
#define ABS(a)		(((a)<0) ? -(a) : (a))

/* take binary sign of a, either -1, or 1 if >= 0 */
#define SGN(a)		(((a)<0) ? -1 : 1)



/************************ 160x160 1-bit ************************/

static void inline draw_pixel_1bit_160(int x,int y)
{
if(x<0 || y<0 || x>=160 || y>=160) return;
scrnbmp[y*20+(x>>3)]|=(1<<(7-(x&7)));
}

static void draw_line_1bit_160(int x1,int y1,int x2,int y2)
{
int d, x, y, ax, ay, sx, sy, dx, dy;

/* if all endpoints are onscreen, use slightly faster methods
 * (actually, allow for one-pixel border so pointer always
 * points within screen buffer)
 */
if(x1>=1 && x1<159 && y1>=1 && y1<159 &&
   x2>=1 && x2<159 && y2>=1 && y2<159)
  {
  unsigned char *ptr,mask;
  
  dx = x2-x1;  ax = ABS(dx)<<1;  sx = SGN(dx);
  dy = y2-y1;  ay = ABS(dy)<<1;  sy = SGN(dy);

  ptr=scrnbmp+y1*20+(x1>>3);
  mask=(1<<(7-(x1&7)));
  if (ax>ay)
    {		/* x dominant */
    d = ay-(ax>>1);
    x = x1;
    for (;;)
      {
      *ptr|=mask;
      if (x==x2) return;
      if (d>=0)
        {
        if(sy)
          ptr+=(sy>0)?20:-20;
        d -= ax;
        }
      if(sx)
        {
        if(sx>0)
          { mask>>=1; if(!mask) mask=128,ptr++; }
        else
          { mask<<=1; if(!mask) mask=1,ptr--; }
        }
      x += sx;
      d += ay;
      }
    }
  else 
    {			/* y dominant */
    d = ax-(ay>>1);
    y = y1;
    for (;;) 
      {
      *ptr|=mask;
      if (y==y2) return;
      if (d>=0)
        {
        if(sx)
          {
          if(sx>0)
            { mask>>=1; if(!mask) mask=128,ptr++; }
          else
            { mask<<=1; if(!mask) mask=1,ptr--; }
          }
        d -= ay;
        }
      if(sy)
        ptr+=(sy>0)?20:-20;
      y += sy;
      d += ax;
      }
    }

  return;
  }

/* if it's partially offscreen, do it the slowest way. */

dx = x2-x1;  ax = ABS(dx)<<1;  sx = SGN(dx);
dy = y2-y1;  ay = ABS(dy)<<1;  sy = SGN(dy);

x = x1;
y = y1;
if (ax>ay)
  {		/* x dominant */
  d = ay-(ax>>1);
  for (;;)
    {
    draw_pixel_1bit_160(x,y);
    if (x==x2) return;
    if (d>=0)
      {
      y += sy;
      d -= ax;
      }
    x += sx;
    d += ay;
    }
  }
else 
  {			/* y dominant */
  d = ax-(ay>>1);
  for (;;) 
    {
    draw_pixel_1bit_160(x,y);
    if (y==y2) return;
    if (d>=0)
      {
      x += sx;
      d -= ay;
      }
    y += sy;
    d += ax;
    }
  }
}

static void draw_bmp_1bit_160(int x,int y,int w,int h,unsigned char *bmp)
{
unsigned char *ptr;
int oldw=(w>>3);

if(x<0 || y<0 || x+w>160 || y+h>160) return;

ptr=scrnbmp+y*20+(x>>3);
for(;h;h--)
  {
  for(w=oldw;w;w--)
    *ptr++|=*bmp++;
  ptr+=20-oldw;
  }
}


/************************ 160x160 8-bit ************************/

static void inline draw_pixel_8bit_160(int x,int y)
{
if(x<0 || y<0 || x>=160 || y>=160) return;
scrnbmp[y*160+x]=black_8bit;
}

static void draw_line_8bit_160(int x1,int y1,int x2,int y2)
{
int d, x, y, ax, ay, sx, sy, dx, dy;

/* if all endpoints are onscreen, use slightly faster methods
 * (actually, allow for one-pixel border so pointer always
 * points within screen buffer)
 */
if(x1>=1 && x1<159 && y1>=1 && y1<159 &&
   x2>=1 && x2<159 && y2>=1 && y2<159)
  {
  unsigned char *ptr;
  
  dx = x2-x1;  ax = ABS(dx)<<1;  sx = SGN(dx);
  dy = y2-y1;  ay = ABS(dy)<<1;  sy = SGN(dy);

  ptr=scrnbmp+y1*160+x1;
  if (ax>ay)
    {		/* x dominant */
    d = ay-(ax>>1);
    x = x1;
    for (;;)
      {
      *ptr=black_8bit;
      if (x==x2) return;
      if (d>=0)
        {
        if(sy)
          ptr+=(sy>0)?160:-160;
        d -= ax;
        }
      if(sx)
        {
        if(sx>0)
          ptr++;
        else
          ptr--;
        }
      x += sx;
      d += ay;
      }
    }
  else 
    {			/* y dominant */
    d = ax-(ay>>1);
    y = y1;
    for (;;) 
      {
      *ptr=black_8bit;
      if (y==y2) return;
      if (d>=0)
        {
        if(sx)
          {
          if(sx>0)
            ptr++;
          else
            ptr--;
          }
        d -= ay;
        }
      if(sy)
        ptr+=(sy>0)?160:-160;
      y += sy;
      d += ax;
      }
    }

  return;
  }

/* if it's partially offscreen, do it the slowest way. */

dx = x2-x1;  ax = ABS(dx)<<1;  sx = SGN(dx);
dy = y2-y1;  ay = ABS(dy)<<1;  sy = SGN(dy);

x = x1;
y = y1;
if (ax>ay)
  {		/* x dominant */
  d = ay-(ax>>1);
  for (;;)
    {
    draw_pixel_8bit_160(x,y);
    if (x==x2) return;
    if (d>=0)
      {
      y += sy;
      d -= ax;
      }
    x += sx;
    d += ay;
    }
  }
else 
  {			/* y dominant */
  d = ax-(ay>>1);
  for (;;) 
    {
    draw_pixel_8bit_160(x,y);
    if (y==y2) return;
    if (d>=0)
      {
      x += sx;
      d -= ay;
      }
    y += sy;
    d += ax;
    }
  }
}

static void draw_bmp_8bit_160(int x,int y,int w,int h,unsigned char *bmp)
{
unsigned char *ptr;
int oldw=w,mask;

if(x<0 || y<0 || x+w>160 || y+h>160) return;

ptr=scrnbmp+y*160+x;
for(;h;h--)
  {
  mask=128;
  for(w=oldw;w;w--)
    {
    if(mask&*bmp)
      *ptr=black_8bit;
    ptr++;
    mask>>=1;
    if(!mask)
      bmp++,mask=128;
    }
  ptr+=160-oldw;
  }
}


/************************ fallback ************************/

/* these are *HORRIBLY* slow and only used if we absolutely have to. */

static void draw_pixel_fallback(int x,int y)
{
/* since pixel stuff isn't *that* important in Hyperoid, we
 * don't bother with this when using the OS. It's slow enough
 * like that already. :-/
 */
}

static void draw_line_fallback(int x1,int y1,int x2,int y2)
{
WinDrawLine(x1,y1,x2,y2);
}

/* this is painful, but it's hard to see an alternative... */
static void draw_bmp_fallback(int x,int y,int w,int h,unsigned char *bmp)
{
int oldw=w,oldx=x,mask;

for(;h;h--,y++)
  {
  mask=128;
  for(w=oldw,x=oldx;w;w--,x++)
    {
    if(mask&*bmp)
      WinDrawPixel(x,y);
    mask>>=1;
    if(!mask)
      bmp++,mask=128;
    }
  }
}


/************************ general ************************/

/* returns zero if we couldn't get an offscreen buffer.
 * It calls draw_uninit() itself if this happens, so the caller
 * shouldn't do that on error.
 *
 * Note that after a successful draw_init() call, you must
 * call draw_select_buffer() before calling anything which does
 * any drawing.
 */
int draw_init(void)
{
UInt16 wcowerr;
UInt32 romver;
UInt32 w=160,h=160,d=1;
Boolean encol=false;

FtrGet(sysFtrCreator,sysFtrNumROMVersion,&romver);
if(romver>=0x03003000)
  featureset30=1;
if(romver>=0x03503000)
  featureset35=1;

draw_pixel=draw_pixel_1bit_160;
draw_line=draw_line_1bit_160;
draw_bmp=draw_bmp_1bit_160;
buffersize=160*160/8;

/* OS 2.x doesn't have WinScreenMode() equivalent, but we know it'll
 * be 160x160 1-bit on that anyway.
 */
if(featureset30)
  {
  if(WinScreenMode(winScreenModeGet,&scrwidth,&scrheight,&scrdepth,&screncol)!=0)
    return(0);

  /* try for 160x160 1-bit */
  if(WinScreenMode(winScreenModeSet,&w,&h,&d,&encol)!=0)
    {
    /* if not, try 8-bit colour */
    d=8; encol=true;
    if(WinScreenMode(winScreenModeSet,&w,&h,&d,&encol)==0)
      {
      /* got it */
      draw_pixel=draw_pixel_8bit_160;
      draw_line=draw_line_8bit_160;
      draw_bmp=draw_bmp_8bit_160;
      buffersize=160*160;
      }
    else
      {
      /* if all else failed and we have OS 3.5, we'll use OS calls
       * (painfully slow, but better than nothing (just about)).
       *
       * I think in practice it's impossible for both mode
       * changes to fail on a pre-3.5 machine, but just in case...
       */
      if(!featureset35)
        return(0);
      
      draw_pixel=draw_pixel_fallback;
      draw_line=draw_line_fallback;
      draw_bmp=draw_bmp_fallback;
      buffersize=0;
      
      draw_is_using_os=1;
      w=scrwidth;
      h=scrheight;
      
      /* what it's set to here doesn't matter too much,
       * we'll be changing it once we start drawing anyway.
       * We're only setting them to get the old values.
       */
      cols_changed=1;
      old_back=WinSetBackColor(0);
      old_fore=WinSetForeColor(0);
      }
    }
  }

if(featureset35)
  {
  RGBColorType cols[8]=
    {
    {0,  0,  0,  0},
    {0,  0,  0,255},
    {0,255,  0,  0},
    {0,255,  0,255},
    {0,  0,255,  0},
    {0,  0,255,255},
    {0,255,255,  0},
    {0,255,255,255}
    };
  int f;
  
  for(f=0;f<8;f++)
    cols_8bit[f]=WinRGBToIndex(cols+f);
  }

if((offscreenwin=WinCreateOffscreenWindow(w,h,screenFormat,&wcowerr))==NULL)
  {
  draw_uninit();
  return(0);
  }

return(1);
}


/* NB: must also be callable from above in case the
 * offscreen win can't be made.
 */
void draw_uninit(void)
{
if(offscreenwin)
  WinDeleteWindow(offscreenwin,false);

if(cols_changed)
  WinSetBackColor(old_back),WinSetForeColor(old_fore);

if(featureset30)
  WinScreenMode(winScreenModeSet,
                &scrwidth,&scrheight,&scrdepth,&screncol);
}


/* inefficient, but we run quickly enough in 1-bit that this
 * shouldn't hurt. (Other depths are dealt with using inverted cols.)
 * Obviously, this should only ever be called when using 1-bit.
 */
static void inline invert_bits(void)
{
unsigned char *ptr;

for(ptr=scrnbmp;ptr<scrnbmp+3200;ptr++)
  *ptr=~*ptr;
}


/* once called, this must not be called again before
 * calling either draw_deselect_buffer() *OR* draw_copy_buffer()
 * (but not both).
 */
void draw_select_buffer(void)
{
static int gotbounds;

if(!gotbounds)
  {
  WinGetClip(&oswbounds);
  gotbounds=1;
  }

currentwin=WinSetDrawWindow(offscreenwin);

/* our notion of white/black can change at `any' time,
 * so redo the colours each time. The "&7" should
 * definitely be unnecessary, but what the heck. :-)
 */
if(prefs.invert)
  white_8bit=cols_8bit[prefs.fg&7],black_8bit=cols_8bit[prefs.bg&7];
else
  white_8bit=cols_8bit[prefs.bg&7],black_8bit=cols_8bit[prefs.fg&7];
if(cols_changed)
  WinSetBackColor(white_8bit),WinSetForeColor(black_8bit);

/* get address of offscreen buffer's bitmap */
if(featureset35)
  scrnbmp=BmpGetBits(WinGetBitmap(offscreenwin));
else
  scrnbmp=offscreenwin->displayAddrV20;

/* when paused we reuse the old buffer rather than clearing it.
 * So when using 1-bit we need to invert it back when inverted.
 */
if(prefs.invert && buffersize==3200 && paused)
  invert_bits();
}


/* deselect offscreen buffer, but don't copy to screen.
 */
void draw_deselect_buffer(void)
{
WinSetDrawWindow(currentwin);
}


/* copy buffer to screen, and deselect.
 */
void draw_copy_to_screen(void)
{
if(prefs.invert && buffersize==3200)
  invert_bits();

WinSetDrawWindow(currentwin);
MenuEraseStatus(NULL);	/* in case graffiti command bit is visible */
WinCopyRectangle(offscreenwin,0,&oswbounds,0,0,winPaint);
}


/* erase contents of the buffer (must be selected) */
void draw_clear_buffer(void)
{
if(!buffersize)
  {
  WinEraseWindow();
  return;
  }

if(buffersize==3200)
  memset(scrnbmp,0,buffersize);
else
  memset(scrnbmp,white_8bit,buffersize);
}
